; Pic Counter
; Version: 1.2
; This version for the PIC16F628 (and probably PIC16F627)
; Clock frequency 10.240000MHz 
; Instruction cycle time = 0.390625 microseconds. (1/10.24)*4

;This code is licenced under the GNU GENERAL PUBLIC LICENSE version 2. 
;See the text file 'COPYING' for details.

;*********************************************************************
;Changes  
; First release 21 Jul 2006

;10 Oct. 2006
;Fix annoying 'not in bank0' warning message.
;Slight mod to avoid PIC interrupt vector (0x05).

;12 Oct. 2006
;Increase delay after CLEAR and HOME. To allow sluggish- 
;display modules to catch up
;*********************************************************************



	ERRORLEVEL -302		;Get rid of annoying 'not in bank0' warning
	LIST	P=16f628A
		
	__CONFIG	_EXTCLK_OSC & _PWRTE_ON & _WDT_OFF & _LVP_OFF
	
	include p16f628A.inc
	
count1	equ	0x20
count2	equ	0x21
count3	equ	0x22
count4	equ	0x23
byte0	equ	0x24
byte1	equ	0x25
byte2	equ	0x26
byte3	equ	0x27
r0	equ	0x28
r1	equ	0x29
r2	equ	0x2a
r3	equ	0x2b
r4	equ	0x2c
chr	equ	0x2d
temp	equ	0x2e
count	equ	0x2f
rs	equ	1
ena	equ	0
LED	equ	2
gate	equ	3




;******** MAIN PROGRAM *********

	org	0x00		;Start here.
	goto	Start
	
Start	org	0x05
	movlw	0x07
	movwf	CMCON		;Comparator - i/o pin setup 	
	bsf	3,5		;Switch to f register page 1 	
	movlw	0xf4		;Configure PIC I/O pins
	movwf	PORTA		;PORTA 0 and 1 outputs
	movlw	0x0f		;PORTB 4 to 7 = outputs			
	movwf	PORTB

	movlw	0x67
	movwf	OPTION_REG	;Prescaler /256, PORTB pull up dis.
	bcf	3,5		;Switch to f register page 0	
	clrf	PORTA
	clrf	PORTB
	
	
	

	
	call	lcdinit		;Initialise LCD module

	movlw	'C'		; Start up message
	call	display 
	movlw	'O'
	call	display 
	movlw	'U'
	call	display
	movlw	'N'
	call	display
	movlw	'T'
	call	display
	movlw	'E'
	call	display
	movlw	'R'
	call	display
	movlw	' '
	call	display
	movlw	' '
	call	display
	movlw	'V'
	call	display
	movlw	'e'
	call	display
	movlw	'r'
	call	display
	movlw	' '
	call	display
	movlw	'1'
	call	display
	movlw	'.'
	call	display
	movlw	'2'
	call	display

	
	movlw	0x19		;delay for 2.5 sec.	
	call	deci
	call	clear	
	
	call	line2
	
	movlw	'G'
	call	display
	movlw	'T'
	call	display
	movlw	'='
	call	display
	movlw	'1'
	call	display
	movlw	'.'
	call	display
	movlw	'0'
	call	display
	movlw	' '
	call	display
	movlw	' '
	call	display
	movlw	'P'
	call	display
	movlw	'S'
	call	display
	movlw	'='
	call	display
	movlw	'1'
	call	display
	movlw	':'
	call	display
	movlw	'1'
	call	display
	call	home
		

;********* COUNTER RESET ***********	
	
;** The block of code below stops the counter by closing the counter gate - 
;** and resets the PICs counter/timer. It also resets the registers that  -
;** used to store the result of the previous count

	bcf	PORTA,gate	;Counter gate pin (RA3) low.
	bsf	3,5		;Switch to f register page 1
	bcf	PORTA,gate	;Make counter gate pin (RA3) an output (close gate)
	bcf	3,5		;Back to page 0
	
cnt_agn	clrf	byte3		;Reset count registers. This resets the -
	clrf	byte2		;counter to zero before the next count begins.
	clrf	byte1
	clrf	byte0
	
	bcf	INTCON,T0IF	;Reset timer0 overflow/interrupt flag. This flag -
				;shows when the timer/counter overflows. 
				
	clrf	TMR0		;Reset timer0 (including prescaler)
	
 
	
;********* THE COUNT **********
;** This is where the main business is done. 	
	

;** The next three lines open the counter gate by changing  
;** the gate pin (RA,3) from an output to an input.
 
	bsf	3,5		;Switch to f register page 1
	bsf	PORTA,gate	;open gate, start counting ......
	bcf	3,5		;Back to page 0	
	
;** Now we must wait for exactly one second
;** The subroutine 'onesec' is a one second delay. Note that this 
;** subroutine also keeps track of counter/timer overflows. 

	call	onesec 
	
	bsf	3,5		;Switch to f register page 1
	bcf	PORTA,gate	;Close gate at precisely one second.
	bcf	3,5		;Back to page 0	
	
;** The count is now complete. 
;********************************************


;*** A FEW NOTES ********
;*
;* The gate is now closed and the count is frozen. The count is 
;* stored as a 32 bit number in four separate eight bit registers. 
;* The most significant 16 bits are stored in the register: byte2 
;* and byte3.  The least significant 16 bits are stored in the 
;* TMR0 register and the 1:256 prescaler. The value in TMR0 can 
;* be read directly. The prescaler can not be read directly, so 
;* we will need to apply a little bit of trickery to recover this 
;* value.


	movf	TMR0,w		;Get the value in TMR0
	movwf	byte1		;Store it
	
                		;Now for the trickery	
				
	clrf	temp
	
clk_agn	incf	temp,f          ;Increase pulse count
	bsf	PORTA,gate	;Send clock pulse to TMR0 -
	nop			; input (pin 3)
	bcf	PORTA,gate
	movf	TMR0,w		;Get the value in TMR0
	subwf	byte1,w		;Check if it has changed as a 
	btfsc	STATUS,Z	;result of the clock pulse
	goto	clk_agn		;No change, try again......
	comf	temp,f		;Yes, the prescaler has overflowed.
	incf	temp,w		;Negate the pulse count and -		
	movwf	byte0		;store it in byte0
	
;* The full 32 bit count result is now stored in 
;* byte0, byte1, byte2 and byte3. 


;* The number needs to be converted to BCD before it is 
;* sent to the LCD display module. 	
	
done	call	b2_bcd		;Convert count to bcd 	
	
	call	home		;'Home' display before writing. 
	
;* The code below sends the packed BCD numbers to the 
;* LCD display one digit at at time. Note that the display 
;* is in 4 bit mode so each number is sent as two 4 bit nibbles. 
	
	movlw	0x30		
	movwf	temp		;Amount to add to convert to ascii
	
				;As all numbers are BCD (1 to 9),
				;they can be converted to ASCII by 
				;simply adding 30HEX to each digit. 
				
	movf	r4,w		;Check the first digit for a leading zero
	andlw	0x0f
	btfss	STATUS,Z
	goto	dig1		;No LZ, display full freq.
	swapf	r3,w		;Check the second digit for leading zero.
	andlw	0x0f
	btfss	STATUS,Z
	goto	dig2
	goto	dig3
	
dig1	movf	r4,0		;Get first bcd digit
	andlw	0x0f		;Mask off other packed bcd digit'	
	addwf	temp,w		;Convert to ascii (add 0x30)
	call	display		;Display it

dig2	swapf	r3,0		;Get 2nd bcd digit
	andlw	0x0f		;Mask off other packed bcd digit	
	addwf	temp,w		;Convert to ascii (add 0x30)
	call	display		;Display it

dig3	movf	r3,0		;Get other bcd digit
	andlw	0x0f		;Mask off other packed bcd digit
	addwf	temp,w		;Convert to ascii (add 0x30)
	call	display		;Display it

	movlw	'.'		;Decimal point.
	call	display	
	
	swapf	r2,0		;Get next digit
	andlw	0x0f		;Mask off other packed bcd digit
	addwf	temp,w
	call	display		;Display it
	

	movf	r2,0		;Get other bcd digit
	andlw	0x0f		;Mask off other packed bcd digit
	addwf	temp,w
	call	display

	swapf	r1,0		;Get next digit
	andlw	0x0f		;Mask off other packed bcd digit
	addwf	temp,w
	call	display		;Display it
	
	movlw	','		;Ok John
	call	display		;I'll stick in a comma. 

	movf	r1,0		;Get other bcd digit
	andlw	0x0f
	addwf	temp,w
	call	display
	

	swapf	r0,0		;Get next digit
	andlw	0x0f		;Mask off other packed bcd digit
	addwf	temp,w
	call	display		;Display it
	

	movf	r0,0		;Get other bcd digit
	andlw	0x0f
	addwf	temp,w
	call	display		;Display last digit (1hz)
	
	movlw	' '		;Write MHz after freq.
	call	display
	movlw	'M'
	call	display
	movlw	'H'
	call	display
	movlw	'z'
	call	display
	movlw	' '		
	call	display
	movlw	' '
	call	display
	
	
	call	home


	goto	cnt_agn		;all done , start again	
				
					


	
	
;******** DELAY LOOP 1 SECOND ***********
onesec	movlw	0x07			;1S delay loop		
	movwf	count1 		              
a1	movlw	0x7e                   
	movwf	count2                  
a2	movlw	0xb5
	movwf	count3 
a3	call	ovcount			;Count number of rtcc overflows                 
	decfsz	count3,1
	goto	a3
	decfsz	count2,1
	goto	a2
	decfsz	count1,1
	goto	a1
	movlw	0x87
	movwf	count1
a4	call	ovcount
	decfsz	count1,1
	goto	a4
	nop
	nop
	nop
        return
;****************************************



;********* COUNT RTCC OVERFLOWS *** 11 CYCLES  ***********
ovcount	btfsc	INTCON,T0IF	; RTCC overflow flag
	goto	ov1
	nop			; No overflow.
	nop			; Lets waste some time.......
	nop
	nop
	nop
	nop
	nop	
	return	
	
ov1	incf	byte2,f	; RTCC overflow, add 1 to highest byte
	btfsc	STATUS,Z	; Has h_byte overflowed
	goto	ov2
	bcf	INTCON,T0IF	; Reset flag
	nop
	nop
	return
	
ov2	incf	byte3,f	; Add 1 to highest byte
	bcf	INTCON,T0IF
	return
	
;********************************************************************

	
	
	
	
;******** DELAY LOOP ********
; **** Calls ms100 **********

deci	movwf	count4		;Delay 100mS x contents of W
nxt6	call	ms100
	decfsz	count4,1
	goto	nxt6
	return
;*****************************	
	
	
	
;******** DELAY LOOP 100MS ***********
ms100	movlw	0x09			;100 mS delay loop		
	movwf	count1            	;including call/return
l1	movlw	0x50                    ;10.240MHz
	movwf	count2                  
l2	movlw	0x75
	movwf	count3                  
l3	decfsz	count3,1
	goto	l3
	decfsz	count2,1
	goto	l2
	decfsz	count1,1
	goto	l1
	movlw	0x77
	movwf	count1
lp99	decfsz	count1,1
	goto	lp99
	nop
        return
;****************************************	



;******** DELAY LOOP 200 MICRO SEC. *************
delay	movlw   0xa9 		; Delay loop 200 microsec
        movwf   count1     	; including call/return
nxt5    decfsz  count1,1	; 10.240MHz
        goto    nxt5   
        return
;*************************************************

;*********** DELAY LOOP 2mS ***********************
ms2	movlw	0x0c 		; Delay 2mS
	movwf	count1
lp15	movlw	0x8c
	movwf	count2
lp16	decfsz	count2,1
	goto	lp16
	decfsz	count1,1
	goto	lp15
	return           
;**************************************************
	

	
;*********** INITIALISE LCD MODULE 4 BIT MODE *******************
lcdinit	call	ms100		;wait for lcd module hardware reset
	bcf	PORTA,rs		;register select low
	bcf	PORTA,ena		;enable line low
	
	movlw	0x30
	movwf	PORTB		;8 bit mode
	bsf	PORTA,ena       ;ena high
	nop
	nop                     ;more than 470 ns
	bcf	PORTA,ena	;ena low
	call	ms100       	;wait for display to catch up


	movlw	0x30
	movwf	PORTB		;8 bit mode
	bsf	PORTA,ena       ;ena high
	nop
	nop                     ;more than 470 ns
	bcf	PORTA,ena	;ena	low
	call	delay       	;wait for display to catch up

	movlw	0x30
	movwf	PORTB		;8 bit mode
	bsf	PORTA,ena       ;ena high
	nop                     ;more than 470 ns
	nop
	bcf	PORTA,ena	;ena	low
	call	delay       	;wait for display to catch up


	movlw	0x20
	movwf	PORTB		;4 bit mode
	bsf	PORTA,ena       ;ena high
	nop                     ;more than 470 ns
	nop
	bcf	PORTA,ena	;ena	low
	call	delay       	;wait for display

	movlw	0x0c		;display on
	call	comd
	call	delay		
	
	movlw	0x06		;entry mode set
	call	comd
	call	delay

	movlw	0x01		;clear display
	call	comd
	call	ms2
	return
;**********************************************************
	
	
;******* comd/display  ******************
; Sends data to LCD display (4 BIT MODE)   

comd	bcf	PORTA,rs	;Send command to LCD module
	goto	cm
display	bsf	PORTA,rs	;Send character to LCD module
cm	movwf	chr		;store char to display
	andlw	0xf0		;mask off lower 4 bits
	movwf	PORTB		;send data to display		
	bsf	PORTA,ena	;ena high
	nop			
	bcf	PORTA,ena	;ena low 

	movlw   0xaa		;delay loop
        movwf   count1     
dl3    	decfsz  count1,1
        goto    dl3    

	movf	chr,0		;get char again 
	andlw	0xf0		;mask off lower 4 bits
	swapf	chr,0           ;swap upper and lower nibbles (4 bit mode)
	movwf	PORTB		;send data to display	
	bsf	PORTA,ena	;ena high
	nop			
	bcf	PORTA,ena	;ena low

	movlw   0xaa 		;delay loop
        movwf   count1     
dl4    	decfsz  count1,1
        goto    dl4      
	return	
;********************************************************************

;************ MOVE TO START OF LINE 2 *****************
line2	movlw	0xc0		;Address for second line of display
	call	comd
	call	delay
	return
;******************************************************

              

;************ CLEAR DISPLAY ***************************

clear	movlw	0x01		;Command to clear display
	call	comd
	call	ms2		;Longer delay needed when clearing display
	call	ms2
	return
;******************************************************	


;*********** MOVE TO HOME *****************************
home	movlw	0x02		;Command to 'home' display.
	call	comd
	call	ms2
	call	ms2
	return
;******************************************************


;******* CONVERT 24 BIT BIN TO PACKED BCD **********

b2_bcd	bcf	STATUS,C
	movlw	0x20
	movwf	count
	clrf	r0
	clrf	r1
	clrf	r2
	clrf	r3
	clrf	r4
loop16	rlf	byte0,f
	rlf	byte1,f
	rlf	byte2,f
	rlf	byte3,f
	rlf	r0,f
	rlf	r1,f
	rlf	r2,f
	rlf	r3,f
	rlf	r4,f

	decfsz	count,1
	goto	adjdec
	retlw	0 

adjdec	movlw	r4
	movwf	FSR 
	call	adjbcd
	
	movlw	r3
	movwf	FSR 
	call	adjbcd

	movlw	r2
	movwf	FSR 
	call	adjbcd

	movlw	r1
	movwf	FSR
	call	adjbcd

	movlw	r0
	movwf	FSR
	call	adjbcd

	goto	loop16

adjbcd	movlw	0x03
	addwf	0,w
	movwf	temp
	btfsc	temp,3
	movwf	0
	movlw	0x30
	addwf	0,w
	movwf	temp
	btfsc	temp,7
	movwf	0
	return	

;****************************************
	
	
	end
